"""Utils dependencies"""
from corkit.utils import FITS, fixwrap, c2_warp, c3_warp, reduce_std_size, save, \
adjust_hdr, get_roll_or_xy, get_sun_center, rotate, rot, get_sec_pixel, get_solar_radius,\
reduce_statistics2, datetime_interval, get_exp_factor, correct_var, DEFAULT_SAVE_DIR

"""Pypi dependencies"""
from datetime import datetime, timedelta
from astropy.io import fits
from typing import Union
from io import BytesIO
import numpy as np
import aiofiles
import aiohttp
import asyncio
import os
import glob

version = '1.0.6'

#done
def level_1(
        fits_file: Union[str, BytesIO],
        target_path: str,
        **kwargs
) -> None:
    filetype = target_path.split('.')[-1]
    #Import data
    print('Importing data')
    img0, header = FITS(fits_file)
    detector = header['detector'].strip().upper()
    header.add_history(f'corkit/lasco.py level_1: (function) {version}, 12/04/24')
    assert (detector in ['C2', 'C3']), f'Not valid detector, not implemented for {detector}'
    #Applying pixelwise implemented corrections
    xsumming = np.maximum(header['sumcol'],1)*np.maximum(header['lebxsum'],1)
    ysumming = np.maximum(header['sumrow'],1)*np.maximum(header['lebysum'],1)
    summing = xsumming*ysumming
    if summing > 1:
        img0 = fixwrap(img0)
        dofull = False
    else:
        dofull = True
    if header['r2col']-header['r1col']+header['r2row']-header['r1row']-1023-1023 != 0: reduce_std_size(img0, header, FULL = dofull)
    
    print(f'LASCO-{header["detector"]}:{header["filename"]}:{header["date-obs"]}T{header["time-obs"]}...')
    match detector:
        case 'C2':
            b, header = c2_calibrate(img0, header, **kwargs)
            b, header = c2_warp(b, header)
            zz = np.where(img0<=0)
            maskall = np.ones((header['naxis1'], header['naxis2']))
            maskall[zz] = 0
            maskall, _ = c2_warp(maskall, header) 
            b*=maskall
        case 'C3':
            b, header = c3_calibrate(img0, header, **kwargs)
            bn, header = c3_warp(b, header)
            zz = np.where(img0<=0)
            maskall = np.ones((header['naxis1'], header['naxis2']))
            maskall[zz]=0
            maskallw, _ = c3_warp(maskall, header)
            _, mask = read_mask_full(header)
            b = bn*maskallw*mask

    img, header = final_step(target_path, filetype, b, header, xsumming, ysumming, **kwargs)

    save(target_path, filetype, img, header)

    return img, header
#done
def final_step(target_path, filetype, img, header, xsumming, ysumming, **kwargs):
    tcr = adjust_hdr(header)
    if header['date'] == '':
        r = get_roll_or_xy(header, 'ROLL', DEGREES=True)
        c = get_sun_center(header, FULL= 1024, MEDIAN=True) 
        cx = c['xcen']
        cy = c['ycen']
    else:
        r = tcr['roll']
        cx = tcr['xpos']
        cy = tcr['ypos']
    if np.abs(header['crota1']) > 170:
        rectify = 180
        cntr = 511.5
        x = cx - cntr
        y = cy - cntr
        cx = cntr + x*np.cos(rectify*np.pi/180.) - y*np.sin(rectify*np.pi/180.)
        cy = cntr + x*np.sin(rectify*np.pi/180.) + y*np.cos(rectify*np.pi/180.)
        img = rotate(img, 2)
        r-=180
    else: rectify=0
    xc = (cx - header['r1col']+ 20)/xsumming
    yc = (cy - header['r1row']+ 1)/ysumming

    r_hdr = r
    if r<-180: r_hdr +=360
    crpix_x = xc+1		
    crpix_y = yc+1

    if 'NOROLL_CORRECTION' in kwargs or np.abs(r)<1:
        pass
    else:
        img = rot(img, -1*r, 1, xc, yc, INTERP = True, PIVOT = True, MISSING=0) 
        rectify+=r
    if filetype == 'fits':
        #Adding final keywords and history modification
        header.add_history(f'CORKIT Level 1 calibration with python modules: to_level1.py, open source level 1 implementation.')
        header['date'] = datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")
        header['filename'] = os.path.basename(target_path)
        header['CRPIX1'] = crpix_x
        header['CRPIX2'] = crpix_y
        header['CROTA'] = r_hdr
        header['CROTA1'] = r_hdr
        header['CROTA2'] = r_hdr
        header['CRVAL1'] = 0
        header['CRVAL2'] = 0
        header['CTYPE1'] = 'HPLN-TAN'
        header['CTYPE2'] = 'HPLN-TAN'
        header['CUNIT1'] = 'arcsec'
        header['CUNIT2'] = 'arcsec'
        platescl = get_sec_pixel(header)
        header['CDELT1'] = platescl
        header['CDELT2'] = platescl
        header['XCEN'] = 0 + platescl*((header['naxis1']+1)/2 - crpix_x)
        header['YCEN'] = 0 + platescl*((header['naxis2']+1)/2 - crpix_y)
        header['DATE-OBS'] = header['DATE-OBS'] + 'T' + header['TIME-OBS']
        header['TIME-OBS'] = ''
        rsun = get_solar_radius(header)
        header['RSUN'] = rsun
    if 'NOSCALE' in kwargs:
        pass
    else:
        #mirar
        if header['detector'].strip().upper() == 'C3' and header['filter'] == 'Clear':
            scalemin = 0
            scalemax = 6.5e-9
            if not 'NOSTAT' in kwargs:
                header = reduce_statistics2(img, header, satmax = scalemax)

            bscale = (scalemax-scalemin)/65536
            bzero=bscale*32769
            ind = np.where(img!=0)
            bout = np.zeros((header['naxis2'], header['naxis1']))
            bout[ind] = np.round((np.maximum(np.minimum(img[ind], scalemax), scalemin) - bzero) / bscale).astype(int)
            if filetype == 'fits':
                header['BSCALE'] = bscale
                header['BZERO'] = bzero
                header['BLANK'] = -32768
                header['COMMENT'] = f' FITS coordinate for center of full image is (512.5,512.5). Rotate image CROTA degrees CCW to correct. Data is scaled between {scalemin} and {scalemax}. Percentile values are before scaling.'
        else:
            if not 'NOSTAT' in kwargs: 
                header = reduce_statistics2(img, header)
            bout = np.float32(img)

    return bout, header
#done
class downloader:
        tools = ['c2', 'c3']
        batch_size = 2
        def __init__(self, tool: str, root: str = './SOHO/LASCO/'):
            assert (tool in self.tools), f'Not in tools: {self.tools}'
            self.tool = tool
            self.lasco_root = root
            self.fits_root = lambda day, hour: os.path.join(self.lasco_root, f'{self.tool}/{day}_{hour}.fits')
            self.url = lambda date, name: f'https://lasco-www.nrl.navy.mil/lz/level_05/{date[2:]}/{self.tool}/{name}'
            self.url_img_txt = lambda date: f'https://lasco-www.nrl.navy.mil/lz/level_05/{date[2:]}/{self.tool}/img_hdr.txt'
            os.makedirs(os.path.join(self.lasco_root, self.tool), exist_ok=True)

        def get_check_tasks(self, scrap_date: tuple[datetime, datetime]):
            scrap_date = datetime_interval(scrap_date[0], scrap_date[-1], timedelta(days = 1))
            self.new_scrap_date_list = [date for date in scrap_date if glob.glob(self.fits_root(date, '*')) == []]

        async def get_download_tasks(self):
            for i in range(0, len(self.new_scrap_date_list), self.batch_size):
                await asyncio.gather(*[self.download_day(day) for day in self.new_scrap_date_list[i:i+self.batch_size]])

        async def download_day(self, day):
            names_hours = await asyncio.gather(self.scrap_metadata(self.url_img_txt(day)))
            await asyncio.gather(*[self.download_url(name, day, hour) for c in names_hours for name, hour in c.items()])

        async def download_url(self, name, day, hour):
            async with aiohttp.ClientSession() as session:
                async with session.get(self.url(day, name), ssl = False) as response, aiofiles.open(self.fits_root(day, hour), 'wb') as f:
                    await f.write(await response.read())

        async def scrap_metadata(self, url):
            async with aiohttp.ClientSession() as session:
                async with session.get(url, ssl = False) as response:
                    data = await response.text()
                    return {line.split()[0]: line.split()[2].replace(':','') for line in data.split('\n')[:-1]}
            
        async def downloader_pipeline(self, scrap_date):
            self.get_check_tasks(scrap_date)
            await self.get_download_tasks()

        async def __call__(self, scrap_date_list):
            for scrap_date in scrap_date_list:    
                await self.downloader_pipeline(scrap_date)
#done
def read_bkg_full():
    bkg_path = os.path.join(DEFAULT_SAVE_DIR, '3m_clcl_all.fts')
    with fits.open(bkg_path) as hdul:
        bkg = hdul[0].data.astype(float)
        bkg *= 0.8/hdul[0].header['exptime']
    return bkg
#done
def read_ramp_full(header):
    ramp_path = os.path.join(DEFAULT_SAVE_DIR,'C3ramp.fts')
    ramp = fits.getdata(ramp_path)
    header.add_history('C3ramp.fts, 1999/03/18')
    return header, ramp
#done
def read_mask_full(header):
    msk_fn=os.path.join(DEFAULT_SAVE_DIR,'c3_cl_mask_lvl1.fts')
    mask = fits.getdata(msk_fn)
    header.add_history('c3_cl_mask_lvl1.fts 2005/08/08')
    return header, mask
#done
def read_vig_full(date, header):
    if date<51000:
        vig_path = os.path.join(DEFAULT_SAVE_DIR,'c3vig_preint_final.fts')
        header.add_history('c3vig_preint_final.fts')
    else:
        vig_path = os.path.join(DEFAULT_SAVE_DIR,'c3vig_postint_final.fts')
        header.add_history('c3vig_postint_final.fts')
    vig = fits.getdata(vig_path)
    return header, vig
#done
def c3_calibrate(
        img0: np.array, 
        header: dict,
        **kwargs):
    
    assert (header['detector'] == 'C3'), 'Not valid C3 fits file'
    header.add_history('corkit/lasco.py c3_calibrate: (function) - Python implementation')
    #returns the date
    mjd = header['mid_date']
    
    # Get exposure factor and bias
    header, expfac, bias = get_exp_factor(header) # define get_exp_factor
    
    # Correct the raw values
    header['exptime'] *= expfac
    header['offset'] = bias

    # Get calibration factor
    if not 'NO_CALFAC' in kwargs:
        header, calfac = c3_calfactor(header)
    else:
        calfac = 1.0
        header.add_history('No calibration factor: 1')

    # Get mask, ramp, bkg(fuzzy) and vignetting function
    header, vig = read_vig_full(mjd, header)
    header, mask = read_mask_full(header)
    header, ramp = read_ramp_full(header)
    bkg = read_bkg_full()

    if not 'NO_VIG' in kwargs:
        pass
    else:
        vig = np.ones_like(vig)
    
    vig, ramp, bkg, mask = correct_var(header, vig, ramp, bkg, mask)

    img = c3_calibration_forward(img0, header, calfac, vig, mask, bkg, ramp, **kwargs)
    
    header.add_history(f'corkit/lasco.py c3_calibrate: (function) {version}, 12/04/24')

    return img, header
#done
def c3_calibration_forward(
        img0: np.array,
        header,
        calfac: float,
        vig: np.array,
        mask: np.array,
        bkg: np.array,  
        ramp: np.array,
        **kwargs,
):
    
    
    if header['fileorig'] == 0:
        img = img0/header['exptime']
        img = img*calfac*vig - ramp
        if not 'NO_MASK' in kwargs.keys(): img*=mask
        return img
    
    if header['filter'] != 'Clear': ramp = 0

    header['polar'] = header['polar'].strip()

    if header['polar'] ==  'PB' or \
            header['polar'] == 'TI' or \
            header['polar'] == 'UP' or \
            header['polar'] == 'JY' or \
            header['polar'] == 'JZ' or \
            header['polar'] == 'Qs' or \
            header['polar'] == 'Us' or \
            header['polar'] == 'Qt' or \
            header['polar'] == 'Qt' or \
            header['polar'] == 'Jr' or \
            header['polar'] == 'Jt':
        
        img = img0/header['exptime']
        img = img*calfac*vig
        
        if not 'NO_MASK' in kwargs: img*=mask
        
        return img
    else:
        img = (img0-header['offset'])/header['exptime']
        ## recons
        img = img*vig*calfac - ramp
        if not 'NO_MASK' in kwargs:
            img*=mask
        return img
#done
def c3_calfactor(header,**kwargs):
    # Set calibration factor for the various filters
    filter_ = header['filter'].upper().strip()
    polarizer = header['polar'].upper().strip()
    mjd = header['mid_date']
    if filter_ == 'ORANGE':
        cal_factor = 0.0297
        polref = cal_factor / 0.25256  # absolute value of +60
        if polarizer == 'CLEAR':
            cal_factor *= 1.0
        elif polarizer == '+60DEG':
            cal_factor = polref
        elif polarizer == '0DEG':
            cal_factor = polref * 0.9648
        elif polarizer == '-60DEG':
            cal_factor = polref * 1.0798
        else:
             cal_factor*=1
    elif filter_ == 'BLUE':
        cal_factor = 0.0975
        polref = cal_factor / 0.25256  # absolute value of +60
        if polarizer == 'CLEAR':
            cal_factor *= 1.0
        elif polarizer == '+60DEG':
            cal_factor = polref
        elif polarizer == '0DEG':
            cal_factor = polref * 0.9734
        elif polarizer == '-60DEG':
            cal_factor = polref * 1.0613
        else:
             cal_factor*=1
    elif filter_ == 'CLEAR':
        cal_factor = 7.43e-8 * (mjd - 50000) + 5.96e-3
        polref = cal_factor / 0.25256  # absolute value of +60
        if polarizer == 'CLEAR':
            cal_factor *= 1.0
        elif polarizer == '+60DEG':
            cal_factor = polref
        elif polarizer == '0DEG':
            cal_factor = polref * 0.9832
        elif polarizer == '-60DEG':
            cal_factor = polref * 1.0235
        elif polarizer == 'H_ALPHA':
            cal_factor = 1.541
        else:
             cal_factor = 0.
    elif filter_ == 'DEEPRD':
        cal_factor = 0.0259
        polref = cal_factor / 0.25256  # absolute value of +60
        if polarizer == 'CLEAR':
            cal_factor *= 1.0
        elif polarizer == '+60DEG':
            cal_factor = polref
        elif polarizer == '0DEG':
            cal_factor = polref * 0.9983
        elif polarizer == '-60DEG':
            cal_factor = polref * 1.0300
        else:
             cal_factor*=1
    elif filter_ == 'IR':
        cal_factor = 0.0887
        polref = cal_factor / 0.25256  # absolute value of +60
        if polarizer == 'CLEAR':
            cal_factor *= 1.0
        elif polarizer == '+60DEG':
            cal_factor = polref
        elif polarizer == '0DEG':
            cal_factor = polref * 0.9833
        elif polarizer == '-60DEG':
            cal_factor = polref * 1.0288
        else:
             cal_factor*=1
    else:
        cal_factor = 0.0

    # Correct calibration factor for pixel summation
    if not 'NO_SUM' in kwargs:
        sumcol = header['SUMCOL']
        sumrow = header['SUMROW']
        lebxsum = header['LEBXSUM']
        lebysum = header['LEBYSUM']
        if sumcol > 0: cal_factor /= sumcol
        if sumrow > 0: cal_factor /= sumrow
        if lebxsum > 1: cal_factor /= lebxsum
        if lebysum > 1: cal_factor /= lebysum
    
    cal_factor *= 1e-10

    header.add_history(f'corkit/lasco.py c3_calfactor: (function) 12/04/24: {cal_factor}')

    return header, cal_factor
#done
def c2_calfactor(header, **kwargs):
    mjd = header['mid_date']
    filter_ = header['filter'].strip().upper()
    polarizer = header['polar'].strip().upper()

    cal_factor = 0.0
    deg = ['+60DEG', '0DEG', '-60DEG', 'ND']

    if filter_ == 'ORANGE':
        cal_factor = 4.60403e-07 * mjd + 0.0374116
        polref = cal_factor / 0.25256
        if polarizer == 'CLEAR':
             cal_factor *= 1
        elif polarizer in deg:
            cal_factor = polref
        else:
            cal_factor *= 1
    elif filter_ in ['BLUE', 'DEEPRD']:
        cal_factor = 0.1033
        polref = cal_factor / 0.25256
        if polarizer == 'CLEAR':
             cal_factor *= 1
        elif polarizer in deg:
            cal_factor = polref
        else:
            if filter_ == 'BLUE':
                cal_factor *= 1
            else:
                 cal_factor = 0
    elif filter_ in ['HALPHA', 'LENS']:
        cal_factor = 0.1033
        polref = cal_factor / 0.25256
        if polarizer == 'CLEAR':
             cal_factor *= 1
        elif polarizer in deg:
            cal_factor = polref
        else:
            cal_factor *= 1
    else:
         cal_factor = 0

    if not 'NO_SUM' in kwargs:
        if header['sumcol'] > 0:
            cal_factor /= header['sumcol']
        if header['sumrow'] > 0:
            cal_factor /= header['sumrow']
        if header['lebxsum'] > 1:
            cal_factor /= header['lebxsum']
        if header['lebysum'] > 1:
            cal_factor /= header['lebysum']

    cal_factor *= 1e-10

    header.add_history(f'corkit/lasco.py c2_calfactor: (function) 12/04/24: {cal_factor}')

    return header, cal_factor
#done
def c2_calibrate(img0, header, **kwargs):
    assert (header['detector']=='C2'), "This is not a C2 valid fits file"

    # Get exposure factor and dark current offset
    header, expfac, bias = get_exp_factor(header) #change for python imp

    header['exptime'] *= expfac
    header['offset'] = bias

    # Calculate calibration factor
    
    if not 'NO_CALFAC' in kwargs:
        header, calfac = c2_calfactor(header, **kwargs)
    else:    
        calfac = 1.0

    # Read vignetting function and mask
    vig_fn = os.path.join(DEFAULT_SAVE_DIR, 'c2vig_final.fts')
    vig_full = fits.getdata(vig_fn)
    
    if not 'NO_VIG' in kwargs:
        # Apply mask to vignetting correction
        vig_full[vig_full < 0.0] = 0.0
        vig_full[vig_full > 100.0] = 100.0
    else:
        vig_full = np.ones_like(vig_full)

    vig_full = correct_var(header, vig_full)[0]

    img = c2_calibration_forward(img0, header, calfac, vig_full)

    header.add_history(f'corkit/lasco.py c2_calibrate: (function) {version}, 12/04/24')

    return img, header
#done
def c2_calibration_forward(img0, header, calfac, vig):
    if header['polar'] in ['PB', 'TI', 'UP', 'JY', 'JZ', 'Qs', 'Us', 'Qt', 'Qt', 'Jr', 'Jt']:
        img = img0/header['exptime']
        img = img*calfac
        img = img*vig
        return img
    else:
        img = (img0 - header['offset']) * calfac / header['exptime']
        img = img*vig
        return img



from skimage.restoration import inpaint
from scipy.fft import fft, ifft
import numpy as np

version = '@(#)reconstruction.py	0.1 , 14/03/24 (Deprecated, using reconstruction autoencoders instead)'

def image_reconstruction(img: np.array):
    map_miss_blocks = -np.fix(img > 0.1)

    # Find locs
    mask = np.zeros(img.shape, dtype = bool)

    mask[map_miss_blocks == 0] = 1
    
    # Perform restoration
    img_restored = inpaint.inpaint_biharmonic(img, mask)
    
    return img_restored

def fuzzy_image():
    side = 32
    too_many = 0


def read_zone(img, list_miss_blocks, rebindex):
    # Side of a square block
    side = 32

    # Extremes columns and rows of the missing zone
    i1 = np.min(list_miss_blocks[0, :] - 1 > 0)
    i2 = np.max(list_miss_blocks[0, :] + 1 < ((32 / rebindex) - 1))
    j1 = np.min(list_miss_blocks[1, :] - 1 > 0)
    j2 = np.max(list_miss_blocks[1, :] + 1 < ((32 / rebindex) - 1))

    # Dimensions of the zone (array of pixels)
    zone_left = side * i1
    zone_right = side * (i2 + 1) - 1
    zone_bottom = side * j1
    zone_top = side * (j2 + 1) - 1
    zone_width = zone_right - zone_left + 1
    zone_height = zone_top - zone_bottom + 1

    zone = img[zone_left:zone_right, zone_bottom:zone_top]

    return zone_width, zone_height, zone

def dct(array: np.array, inverse: bool = False):
    shape = array.shape
    dim = len(shape)

    if dim == 1:
        n = int(shape)
        k = np.arange(n)
        w4n = np.exp(k*complex(0, -2*np.pi/(4*n)))

        sub = np.concatenatate((k, np.flip(k)))
        sub = sub[::2]

        if inverse:
            vv = 0.5 / w4n * (array[k] - array[n- k[::-1]])
            v = ifft(vv)
            x = np.zeros(n)
            x[sub] = v
        else:
            v = array[sub]
            vv = fft(v)
            x = 2*np.real(w4n*vv[k])

    elif dim == 2:
        n2, n1 = shape

        k1 = np.arange(n1) @ np.ones(n2)
        k2 = np.ones(n1) @ np.arange(n2)

        n1_k1 = np.roll(np.flip(k1, 1), 1, axis = 1)
        n2_k2 = np.roll(np.flip(k2, 0), 1, axis = 0)

        sub1 = np.concatenate((k1, np.flip(k1, 1)))
        sub1 = sub1[:, 2*np.arange(n1)]
        sub2 = np.concatenate(([k2], np.flip(k2, 0)))
        sub2 = sub2[2*np.arange(n2), :]

        w4n1 = np.exp(k1*complex(0, -2*np.pi/(4*n1)))
        w4n2 = np.exp(k2*complex(0, -2*np.pi/(4*n2)))

        if inverse:
            mask1 = np.ones(n2,n1)
            mask1[:, 0] = 0
            mask2 = np.ones(n2,n1)
            mask2[0, :] = 0

            vv = 1/4/w4n1/w4n2 * complex(array[k2, k1] - array[n2_k2, n1_k1] * mask1 * mask2, -(array[k2, n1_k1] * mask1 + array[n2_k2, k1]*mask2))
            v = ifft(vv)
            x = np.zeros(n2, n1)
            x[sub2, sub1] = v
        else:
            vv = fft(array[sub2, sub1])
            x = 2*float(w4n1 * (w4n2*vv[k2, k1] + 1/w4n2*vv[n2_k2, k1]))
    
    return x


def fuzzy_block(img, i, j, rebindex):
    side = 32
    zone_width, zone_height, zone = read_zone(img, np.array([i,j]), rebindex)
    
    #given zone: (height, width)
    ny = zone.shape[0]/side 
    nx = zone.shape[1]/side

    sp = np.zeros((side, side, ny, nx), dtype=complex)


    for jj in range(ny):
        for ii in range(nx):
            block = read_block(zone, ii, jj)
            sp[:, :, jj, ii] = dct(block)

    x0 = 0
    y0 = 0
    side = 32
    xx = np.arange(side)
    yy = np.ones(side)  # Equivalent to replicate(1, side) in IDL

    sub_mask = np.zeros((side, side), dtype=np.uint8)

    # Define sub-mask values
    mean, low, horiz, vert, diag, high = 0, 1, 2, 3, 4, 5

    # Parameters for sub-mask generation
    diag_angle = np.pi / 6
    low_radius = 4
    high_radius = side * 0.9

    # Generate sub-mask
    sub_mask[:, :] = diag
    sub_mask[np.abs(yy - y0) < np.abs(xx - x0) * np.tan(diag_angle / 2)] = horiz
    sub_mask[np.abs(xx - x0) < np.abs(yy - y0) * np.tan(diag_angle / 2)] = vert
    sub_mask[np.abs(np.complex(xx - x0, yy - y0)) < low_radius] = low
    sub_mask[np.abs(np.complex(xx - x0, yy - y0)) > high_radius] = high
    sub_mask[y0, x0] = mean

    # Corresponding subspectra calculation
    sub_sp = np.zeros((side, side, 6, ny, nx), dtype=float)
    for jj in range(ny):
        for ii in range(nx):
            for s in range(6):
                sub_sp[:, :, s, jj, ii] = sp[:, :, jj, ii] * (sub_mask == s)

    # Number of elements in each sub-spectrum
    n_sub_sp = np.histogram(sub_mask.astype(int), bins=np.arange(7))[0]

    norm_en = np.zeros(6, ny, nx)

    for jj in range(ny):
        for ii in range(nx):
            for s in range(6):
                norm_en[s, jj,ii] = np.sum(np.abs(sub_sp[:,:,s,jj,ii])**2)/n_sub_sp[s]

    xcent = np.zeros(6, ny, nx)
    ycent = np.zeros(6, ny, nx)

    for jj in range(ny):
        for ii in range(nx):
            for s in range(6):
                xcent[s, jj, ii] = np.sum(xx*np.abs(sub_sp[:,:,s,jj,ii])**2)/np.sum(np.abs(sub_sp[:,:,s,jj,ii])**2)
                ycent[s, jj, ii] = np.sum(yy*np.abs(sub_sp[:,:,s,jj,ii])**2)/np.sum(np.abs(sub_sp[:,:,s,jj,ii])**2)
    
    orient = np.zeros(nx, ny)

    for jj in range(ny):
        for ii in range(nx):
            max_index = np.argmax(norm_en[:, jj, ii] * np.array([0, 0, 1, 1, 1, 0]))
            orient[jj, ii] = max_index
    
    ph_sk = sub_sp < 0

    hist, bins = np.histogram(orient.flatten(), bins=6, range=(0, 5))
    nmax = np.max(hist * np.array([0, 0, 1, 1, 1, 0]))
    smax = np.argmax(hist)

    sim = np.zeros((4, ny, nx))

    for jj in range(ny):
        for ii in range(nx):
            sim[0, jj, ii] = 1 - np.abs(norm_en[smax, jj, ii] - norm_en[smax, 1, 1]) / \
                            np.abs(np.max(norm_en[smax, :, :]) - np.min(norm_en[smax, :, :]))
            
            sim[1, ii, jj] = 1 - np.sqrt((xcent[smax, jj, ii] - xcent[smax, 1, 1])**2 +
                                        (ycent[smax, jj, ii] - ycent[smax, 1, 1])**2) / 20

    member = orient == smax

    member[1,1] = 0

    conf = np.zeros(ny, nx)
    
    for jj in range(ny):
        for ii in range(nx):
            conf[jj, ii] = 1 - np.sum(ph_sk[:, :, smax, jj, ii] ^ ph_sk[:,:, smax, 1,1]) / n_sub_sp(smax)

    all_sub_sp = sub_sp[:,:, smax, :, :]
    all_sub_sp = all_sub_sp[all_sub_sp != 0]
    min_ = np.min(all_sub_sp)
    max_ = np.max(all_sub_sp)

    fuzzy_sub_sp = num_to_fuzzy(sub_sp[:,:,smax,1,1], min_, max_, 0)

    for jj in range(ny):
        for ii in range(nx):
            if member[jj, ii]:
                fuzzy_sub_sp = inter_fuzzy(fuzzy_sub_sp, num_to_fuzzy(sub_sp[:, :, smax, jj, ii], min_, max_, 0.5))
    number_sub_sp = fuzzy_to_num(fuzzy_sub_sp)

    
    sp_block = sp[:, :, 1,1]
    sp_block[sub_mask == smax] = number_sub_sp[sub_mask == smax]
    block = dct(sp_block, inverse = True)

    return block

def num_to_fuzzy(a0, amin, amax, conf):
    # If amin is greater than amax, swap them
    if amax < amin:
        amin, amax = amax, amin
    
    # Ensure each element of a0 is within the range [amin, amax]
    a0 = np.clip(a0, amin, amax)
    
    # Create an array of fuzzy numbers with the same shape as a0
    afuzzy = np.zeros_like(a0, dtype=[('low', float), ('high', float)])
    
    # Assign low and high values for each fuzzy number
    afuzzy['low'] = amin + conf * (a0 - amin)
    afuzzy['high'] = amax - conf * (amax - a0)
    
    return afuzzy

def inter_fuzzy(afuzzy, bfuzzy): #mirar
    cfuzzy = {}
    cfuzzy['low'] = np.maximum(afuzzy['low'], bfuzzy['low'])
    cfuzzy['high'] = np.minimum(afuzzy['high'], bfuzzy['high'])
    return cfuzzy

def fuzzy_to_num(afuzzy):
    a = (afuzzy['low'] + afuzzy['high']) / 2
    return a

def read_block(image, i, j, side=32):
    # Reading a square block from the image
    block = image[side*j:side*(j+1), side*i:side*(i+1)]
    return block

def getl05hdrparam(header):
    out = {}
    out['detector'] = header['detector']
    out['sx'] = header['NAXIS1']
    out['sy'] = header['NAXIS2']
    out['fystart'] = header['R1ROW'] - 1

def get_tmax():
    pass


from corkit.utils import DEFAULT_SAVE_DIR, datetime_interval
from dateutil.relativedelta import relativedelta
from datetime import datetime, timedelta
from collections import defaultdict
from bs4 import BeautifulSoup
from scipy.io import readsav
from itertools import chain
import pandas as pd
import numpy as np
import aiofiles
import aiohttp
import asyncio
import os

#done
def clean_old(file_list):
    dat_files = filter(lambda filename: filename.lower().endswith('.dat'), file_list)
    fts_files = filter(lambda filename: not filename.lower().endswith('.fts'), file_list)
    dat_file = sorted(dat_files)[-1]
    fts_file = sorted(fts_files)[-1]
    return [dat_file, fts_file]
#done
async def get_names(url: str, href):
    async with aiohttp.ClientSession() as client:
        async with client.get(url, ssl = False) as response:
            html = await response.text()
            soup = BeautifulSoup(html, 'html.parser')
            names = soup.find_all('a', href = href)
            names = [name['href'] for name in names]
            return names
#done	
async def download_single(url: str, filepath: str):
    if not os.path.exists(filepath):
        async with aiohttp.ClientSession() as client:
            async with client.get(url, ssl = False) as response, aiofiles.open(filepath, 'wb') as f:
                await f.write(await response.read())
#done
async def update():
    print('Checking calib datasets...')
    base_link = lambda filename: f'https://soho.nascom.nasa.gov/solarsoft/soho/lasco/lasco/data/calib/{filename}'
    base_root = lambda filename: os.path.join(DEFAULT_SAVE_DIR, filename)
    os.makedirs(DEFAULT_SAVE_DIR, exist_ok=True)
    names = [
        'c3vig_postint_final.fts',
        'c3vig_preint_final.fts',
        '3m_clcl_all.fts',
        'C3ramp.fts',
        'c3_cl_mask_lvl1.fts',
        'c2vig_final.fts',
    ]

    links = list(map(base_link, names))
    roots = list(map(base_root, names))

    await asyncio.gather(*[download_single(url, filepath) for url, filepath in zip(links, roots)])
    print('Cheking data_anal directory...')

    os.makedirs(os.path.join(DEFAULT_SAVE_DIR, 'data/data_anal/'), exist_ok=True)
    await download_single('https://soho.nascom.nasa.gov/solarsoft/soho/lasco/lasco/data_anal/data/c2_time_offsets.dat', os.path.join(DEFAULT_SAVE_DIR, 'data/data_anal/c2_time_offsets.dat'))

    print('Checking ancil_data/attitude/predictive dataset...')
    root_base = lambda filename: os.path.join(DEFAULT_SAVE_DIR, f'ancil_data/attitude/predictive/{filename.split("_")[-2][:4]}',filename)

    years = [date[:4] for date in datetime_interval(datetime(2010,1,1), datetime.now(), relativedelta(years = 1))]
    
    for year in years:
        os.makedirs(os.path.join(DEFAULT_SAVE_DIR, 'ancil_data/attitude/predictive', year), exist_ok=True)

    date_base = lambda year: f'{year}/' if int(year)<2022 else ''
    
    pred_link = lambda date: f'https://soho.nascom.nasa.gov/data/ancillary/attitude/predictive/{date_base(date)}'

    names = await asyncio.gather(*[get_names(pred_link(date), lambda href: ('PRE' in href and href.lower().endswith('.fts')) or ('ROL' in href and href.lower().endswith('.dat'))) for date in years])

    names = [*chain.from_iterable(names)]

    cat_str = defaultdict(list)
    
    for filename in names:
        date = filename.split('_')[-2]
        cat_str[date].append(filename)
    
    download_filenames = []
    for file_list in cat_str.values():
        download_filenames.append(clean_old(file_list))
    
    download_filenames = [*chain.from_iterable(download_filenames)]
    
    batch_size = 500
    for i in range(0, len(download_filenames), batch_size):
        await asyncio.gather(*[download_single(pred_link(name.split('_')[-2][:4]) + f'{name}', root_base(name)) for name in download_filenames[i: i+batch_size]])	

    print('Checking nominall_roll_att.dat...')
    os.makedirs(os.path.join(DEFAULT_SAVE_DIR, 'ancil_data/attitude/roll/'), exist_ok=True)
    await download_single('https://soho.nascom.nasa.gov/data/ancillary/attitude/roll/nominal_roll_attitude.dat', os.path.join(DEFAULT_SAVE_DIR, 'ancil_data/attitude/roll/nominal_roll_attitude.dat'))

    print('Checking for .sav files...')
    await medv_update()

    print('Checking for exposure factor files...')
    
    exp_link = lambda date, n_tool: f'https://hesperia.gsfc.nasa.gov/ssw/soho/lasco/lasco/expfac/data/{date[2:6]}/c{n_tool}_expfactor_{date[2:]}.dat'
    exp_root = lambda date, n_tool: os.path.join(DEFAULT_SAVE_DIR, f'data_anal/data/{date[2:6]}/c{n_tool}_expfactor_{date[2:]}.dat') #YYYYMMDD
    
    interval = datetime_interval(datetime(1996, 2, 1), datetime.now(), timedelta(days = 1))
    folders = list(set([date[2:6] for date in interval]))
    for folder in folders:
        os.makedirs(os.path.join(DEFAULT_SAVE_DIR, 'data_anal/data/', folder), exist_ok=True)
    
    batch_size = 250
    for i in range(0, len(interval), batch_size):
        await asyncio.gather(*[download_single(exp_link(date,n_tool), exp_root(date, n_tool)) for date in interval[i: i+batch_size] for n_tool in [2,3]])
    
    print('Checking for occulter_center.dat file...')
    await download_single(
        'https://hesperia.gsfc.nasa.gov/ssw/soho/lasco/idl/convert/occulter_center.dat',
        os.path.join(DEFAULT_SAVE_DIR, 'occulter_center.dat')
    )
#done
async def update_single(url: str, filepath: str):
    if os.path.exists(filepath):
        os.remove(filepath)

    async with aiohttp.ClientSession() as client:
            async with client.get(url, ssl = False) as response, aiofiles.open(filepath, 'wb') as f:
                await f.write(await response.read())
#done
async def medv_update():
    link = lambda file: f'https://soho.nascom.nasa.gov/solarsoft/soho/lasco/idl/data/calib/{file}'
    pathfile = lambda file: os.path.join(DEFAULT_SAVE_DIR, file)
    files = [
        'c2_pre_recovery_adj_xyr_medv2.sav',
        'c3_pre_recovery_adj_xyr_medv2.sav',
        'c2_post_recovery_adj_xyr_medv2.sav',
        'c3_post_recovery_adj_xyr_medv2.sav'
    ]

    await asyncio.gather(*[update_single(link(file), pathfile(file)) for file in files])  
    filepaths = [pathfile(file) for file in files]
    dfs = []
    
    for filepath in filepaths:
        sav_df = readsav(filepath)
        index = pd.DatetimeIndex([date.decode('utf-8') for date in sav_df['c_dt']])
        keys = ['c_r', 'c_rmed', 'c_rmed', 'c_tai', 'c_x', 'c_xmed', 'c_y', 'c_ymed']
        data = {key: sav_df[key] for key in keys}
        new = np.asarray(sav_df['c_utc'])
        data['c_utc_time'] = new['time']
        data['c_utc_mjd'] = new['mjd']
        df = pd.DataFrame(data, index = index, columns = data.keys())
        dfs.append(df)
    
    for filepath, df in zip(filepaths, dfs):
        df.to_csv(filepath)

if __name__ == '__main__':
    asyncio.run(update())